import {
    ACTION_BUTTON_LIST, DONT_SHOW_ACTION_ENTITY_TYPE_LIST, ROUTE_URL
} from 'constant/others/other';
import ENTITY_FORM_LIST from 'constant/templates/form/entity_form_template';
import useDrupalTranslation from 'hooks/drupal_translation';
import MenuTemplate from 'model/MenuTemplate';
import { Button, Form } from 'react-bootstrap';
import { Link } from 'react-router-dom';
import { cspToLabel, entityNameToLabel } from 'service/string';

/**
 * Calculate the list of action buttons to display.
 *
 * @param menuTemplate The template of form menu.
 * @param cloudContext The Cloud Context.
 */
const calcActionButtonList = (menuTemplate: MenuTemplate, cloudContext: string) => {
  const actionTypeList = ENTITY_FORM_LIST.filter((r) => {
    return r.cloudServiceProvider === menuTemplate.cloudServiceProvider
      && r.entityName === menuTemplate.entityName;
  }).map((r) => r.actionType);

  const buttonList = ACTION_BUTTON_LIST
  .filter((r) => {
    return actionTypeList.includes(r.actionType);
  })
  .map((r) => {
    const to = `/${menuTemplate.cloudServiceProvider}/${cloudContext}/${menuTemplate.entityName}/${r.actionType}`;
    const actionLabel = r.label;
    const cloudServiceProviderLabel = cspToLabel(menuTemplate.cloudServiceProvider);
    const entitytype = menuTemplate.entityName;
    return {
      to,
      action: actionLabel,
      cloudServiceProvider: cloudServiceProviderLabel,
      type: entityNameToLabel(entitytype),
    };
  });

  return menuTemplate.entityName === 'instance'
    ? [{
      to: `/${menuTemplate.cloudServiceProvider}/${cloudContext}/server_template`,
      action: 'Add | Launch',
      cloudServiceProvider: cspToLabel(menuTemplate.cloudServiceProvider),
      type: entityNameToLabel('instance'),
    }, ...buttonList]
    : buttonList;
}

/**
 * The button of creating cloud entity.
 *
 * @param menuTemplate The template of form menu.
 * @param cloudContext The Cloud Context.
 */
const ActionButtonGroup = ({ menuTemplate, cloudContext }: {
  menuTemplate: MenuTemplate,
  cloudContext: string
}) => {

  const { t } = useDrupalTranslation();

  /**
   * Entity type that does not display Action buttons in ListView.
   */
  if (DONT_SHOW_ACTION_ENTITY_TYPE_LIST.filter((r) => {
    return r.cloudServiceProvider === menuTemplate.cloudServiceProvider
      && r.entityName === menuTemplate.entityName;
  }).length >= 1) {
    return <></>;
  }

  /**
   * List of action buttons to display.
   */
  const actionButtonList = calcActionButtonList(menuTemplate, cloudContext);

  /**
   * Function to update the entity list.
   */
  const refresh = async () => {
    const url = `/clouds/${menuTemplate.cloudServiceProvider}/${cloudContext}/${menuTemplate.entityName}/update`;

    const response = await fetch(url, { method: 'POST' });
    if (response.ok) {
      window.location.href = `${ROUTE_URL}/${menuTemplate.cloudServiceProvider}/${cloudContext}/${menuTemplate.entityName}`;
    }
  }

  return <Form.Group>
    {
      actionButtonList.map((r) => {
        return <Link className="btn btn-primary" key={r.to} to={r.to}>
          {t(`${r.action} @cloudServiceProvider @entityName`, {
            '@cloudServiceProvider': r.cloudServiceProvider,
            '@entityName': r.type,
          })}
        </Link>;
      })
    }
    <Button onClick={refresh}>
      <span className="glyphicon glyphicon-repeat"></span> {Drupal.t('Refresh')}
    </Button>
  </Form.Group>;

}

export default ActionButtonGroup;
