import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import EntityData from 'model/EntityData';
import { useState } from 'react';
import { useParams } from 'react-router-dom';
import { capitalize, cspToLabel, entityNameToLabel } from 'service/string';

const PageTitle = ({ action, designLabel, bundleId }: {
  action?: string
  designLabel?: string
  bundleId?: string
}) => {
  /**
   * URL parameters.
   */
  const params: {
    cloudServiceProvider?: string,
    cloudContext?: string,
    entityName?: string,
    entityId?: string,
    action?: string
  } = useParams();

  /**
   * Information about the entity specified by the entity ID.
   */
  const [entityData, setEntityData] = useState<EntityData | undefined>(undefined);

  /**
   * Hooks to retrieve entity information.
   */
  const { getJsonData } = useDrupalJsonApi();

  /**
   * The loading entity process in xxUD operation.
   */
  if (entityData === undefined
    && params.cloudServiceProvider !== undefined
    && params.entityName !== undefined
    && params.entityId !== undefined) {

    // Create the URL.
    const entityTypeId = params.entityName === 'cloud_launch_template'
      ? `${params.entityName}` : `${params.cloudServiceProvider}_${params.entityName}`;
    const url_filter = `?filter[drupal_internal__id]=${params.entityId}`;
    const url = params.entityName === 'cloud_launch_template'
      ? `/jsonapi/${entityTypeId}/${params.cloudServiceProvider}${url_filter}`
      : params.entityName === 'namespace_resource_store'
        ? `/jsonapi/cloud_store/${entityTypeId}${url_filter}`
        : `/jsonapi/${entityTypeId}/${entityTypeId}${url_filter}`;

    // Load the entity from the URL.
    getJsonData<{ data: EntityData[] }>(url).then((rawEntityData) => {
      if (rawEntityData.data.length >= 1) {
        setEntityData(rawEntityData.data[0]);
      }
    });
  }

  /**
   * If the existing page title is not displayed,
   * do not display by SPA.
   */
  const pageTitleElements = document.querySelectorAll('h1.page-header');
  if (pageTitleElements.length !== 1) {
    return <></>;
  }
  const pageTitle = pageTitleElements[0];

  /**
   * Get Entity Type Name Label.
   */
  const entityTypeNameLabel = entityNameToLabel(params.entityName ? params.entityName : '');

  /**
   * Get Entity Data Name.
   */
  const entityDataNameLabel = entityData !== undefined
    ? 'name' in entityData.attributes
    ? entityData.attributes['name'] : undefined : undefined;

  /**
   * Get Action Type.
   */
  const actionType = action !== undefined
    ? action : params.action;

  /**
   * Get Action Label.
   */
  const actionLabel = actionType !== undefined
    ? actionType === 'create'
      ? 'Add' : actionType === 'import'
        ? 'Import' : capitalize(actionType) : undefined;

  /**
   * Create new title text.
   */
  let pageTitleText = "Cloud Dashboard";

  // List
  if (params.cloudServiceProvider !== undefined && actionType === undefined) {
    pageTitleText = params.cloudContext === undefined
      ? 'All ' + cspToLabel(params.cloudServiceProvider) + ' ' + entityTypeNameLabel
      : entityTypeNameLabel;
  }

  // Add, Import
  if (actionLabel !== undefined && entityData === undefined) {
    pageTitleText = actionLabel + ' ' + entityTypeNameLabel;
  }

  // Edit
  if (actionLabel !== undefined && entityData !== undefined) {
    pageTitleText = actionLabel + ' ' + entityDataNameLabel;
  }

  // Delete
  if (actionLabel !== undefined && actionType === 'delete') {
    pageTitleText = 'Are you sure you want to delete the ' + entityTypeNameLabel + ' ' + entityDataNameLabel + '?';
  }

  // Detailed View
  if (actionLabel === undefined && entityData !== undefined) {
    pageTitleText = 'Name: ' + entityDataNameLabel;
  }

  // Template List, Project List
  if (designLabel !== undefined) {
    pageTitleText = capitalize(designLabel) + ' list';
  }

  // Store
  if (actionType === undefined && designLabel !== undefined && designLabel === 'store' && bundleId !== undefined) {
    pageTitleText = entityNameToLabel(bundleId);
  }

  /**
   * Rewrite the page title.
   */
  pageTitle.textContent = pageTitleText;

  const siteNameElement =  document.querySelectorAll('#navbar > nav > div.navbar-header > div > a.name.navbar-brand');
  if (siteNameElement.length === 1) {
    const siteName = siteNameElement[0].textContent;
    document.title = pageTitleText + ' | ' + siteName;
  }

  return <></>;

}

export default PageTitle;
