import ENTITY_INFO_TAB_LIST from 'constant/templates/entity/entity_info_tab';
import ENTITY_INFO_LIST from 'constant/templates/entity/entity_info_template';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import useDrupalTranslation from 'hooks/drupal_translation';
import EntityColumn from 'model/EntityColumn';
import EntityData from 'model/EntityData';
import EntityInfoPanelData from 'model/EntityInfoPanelData';
import EntityInfoRecord from 'model/EntityInfoRecord';
import EntityInfoRecordData from 'model/EntityInfoRecordData';
import EntityInfoTemplate from 'model/EntityInfoTemplate';
import DetailActionButtonGroup from 'molecules/DetailActionButtonGroup';
import LoadingSpinner from 'molecules/LoadingSpinner';
import Breadcrumbs from 'organisms/Breadcrumbs';
import EntityInfoPanel from 'organisms/EntityInfoPanel';
import PageTitle from 'organisms/PageTitle';
import SideBar from 'organisms/SideBar';
import { useEffect, useState } from 'react';
import { Col, Container, Row } from 'react-bootstrap';
import { Link, useParams } from 'react-router-dom';
import { convertDataToTextForUI } from 'service/convert';
import { getReplacedUrl } from 'service/string';

/**
 * Custom Hooks for detail page's template.
 */
const useDetailTemplate = () => {

  const params: {
    cloudServiceProvider: string,
    entityName: string,
  } = useParams();

  const entityDetailTemplateList = ENTITY_INFO_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && template.entityName === params.entityName;
  });

  return {
    detailTemplate: entityDetailTemplateList.length >= 1
      ? entityDetailTemplateList[0] : undefined
  };

}

/**
 * Custom Hooks for detail page tab's template.
 */
const useTabTemplate = () => {

  const params: {
    cloudServiceProvider: string,
    entityName: string,
  } = useParams();

  const tabTemplateList = ENTITY_INFO_TAB_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && template.entityName === params.entityName;
  });

  return {
    tabTemplate: tabTemplateList.length >= 1
      ? tabTemplateList[0] : undefined
  };

}

/**
 * Custom Hooks for entity data.
 *
 * @param cloudServiceProvider The cloud service provider.
 * @param entityName The detail entity type ID.
 * @param entityId The entity's ID.
 */
const useEntityData = (
  cloudServiceProvider: string,
  entityName: string,
  entityId: string,
) => {

  const { getJsonData } = useDrupalJsonApi();
  const [entityData, setEntityData] = useState<EntityData | undefined>(undefined);

  useEffect(() => {
    const entityTypeId = `${cloudServiceProvider}_${entityName}`;
    const url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${entityId}`;
    getJsonData<{ data: EntityData[] }>(url).then((result) => {
      if (result.data.length > 0) {
        setEntityData(result.data[0]);
      }
    });
  }, []);

  return {
    entityData
  };

}

/**
 * Custom Hooks for panel data.
 *
 * @param cloudContext The cloud context.
 * @param detailTemplate The detail info template for entity.
 * @param entityId The entity's ID.
 */
const usePanelData = (
  cloudContext: string,
  detailTemplate: EntityInfoTemplate,
  entityId: string
) => {

  const { readDataCache, getJsonData } = useDrupalJsonApi();
  const { entityData } = useEntityData(
    detailTemplate.cloudServiceProvider,
    detailTemplate.entityName,
    entityId
  );
  const [panelDataList, setPanelDataList] = useState<EntityInfoPanelData[]>([]);

  /**
   * Calculate the information on the panel
   * to be displayed from the entity's information.
   *
   * @param dataCache The data cache.
   * @param tableRecordList The record list of table.
   * @param entityColumn The form record's information.
   */
  const getEntityInfoRecordData = async (
    dataCache: Record<string, EntityData[]>,
    tableRecordList: string[],
    entityColumn: EntityColumn
  ) => {
    // Keep the panel from displaying
    // while the entity's information is being read.
    if (!entityData) {
      return {
        type: 'div',
        key: entityColumn.labelName,
        value: ''
      } as EntityInfoRecordData;
    }

    switch (entityColumn.type) {
      case 'metrics': {
        const url = `/clouds/${detailTemplate.cloudServiceProvider}/${entityData.attributes['cloud_context']}/${detailTemplate.entityName}/${entityData.attributes['drupal_internal__id']}/metrics`;
        const metricsData = await getJsonData<Record<string, number>[]>(url);
        return {
          type: 'metrics',
          record: entityColumn.column.map((column) => {
            return {
              title: column.title,
              yLabel: column.yLabel,
              record: metricsData.map((metrics) => {
                return { x: metrics['timestamp'], y: metrics[column.name] };
              })
            }
          })
        } as EntityInfoRecordData;
      }
      case 'custom-table': {
        const keyVal: Record<string, string>[] = [];
        const labelData: Record<string, string> = {};
        entityData.attributes[entityColumn.name].forEach((record: any, index: number) => {
          const temp: Record<string, string> = {};
          for (const column of entityColumn.column) {
            labelData[column.name] = column.labelName;
            const convertedText = convertDataToTextForUI(
              record[column.name],
              column,
              dataCache
            );
            const convertedText2 = column.type === 'link'
              ? getReplacedUrl(
                convertedText,
                entityData.attributes['cloud_context'],
                detailTemplate.entityName,
                entityData.attributes['drupal_internal__id'],
                detailTemplate.cloudServiceProvider,
                index
              ) : convertedText;
            temp[column.name] = convertedText2;
          }
          keyVal.push(temp);
        });
        return {
          type: 'table',
          title: entityColumn.labelName,
          record: keyVal,
          label: labelData
        } as EntityInfoRecordData;
      }
      case 'json-table': {
        const entityColumnList = entityColumn.column;
        const keyVal: Record<string, string>[] = [];

        for (const record of entityData.attributes[entityColumn.name]) {
          const record2: {
            item_key: string,
            item_value: string
          } = record;
          const record3: Record<string, any> = JSON.parse(record2.item_value);

          const temp: Record<string, string> = {};
          for (const entityColumn of entityColumnList) {
            if (entityColumn.name in record3) {
              temp[entityColumn.name] = convertDataToTextForUI(
                record3[entityColumn.name],
                entityColumn,
                dataCache
              );
            }
          }
          keyVal.push(temp);
        }
        return {
          type: 'table',
          title: entityColumn.labelName,
          record: keyVal
        } as EntityInfoRecordData;
      }
      case 'array-table': {
        const keyVal: Record<string, string>[] = [];
        const rows = entityData.attributes[entityColumn.name];
        for (const row of rows) {
          const values: string[] = row.split(',');
          const temp: Record<string, string> = {};
          entityColumn.column.forEach((column, index) => {
            temp[column.name] = convertDataToTextForUI(
              values[index],
              column,
              dataCache
            );
          });
          keyVal.push(temp);
        }
        return {
          type: 'table',
          title: entityColumn.labelName,
          record: keyVal
        } as EntityInfoRecordData;
      }
      default:
        if (tableRecordList.includes(entityColumn.name)) {
          const keyVal: Record<string, string>[] = [];
          for (const record of entityData.attributes[entityColumn.name]) {
            const temp: Record<string, string> = {};
            for (const key of Object.keys((record))) {
              temp[key] = record[key] ? `${record[key].replace(/(\r\n)|\n/, '')}` : '';
            }
            keyVal.push(temp);
          }
          return {
            type: 'table',
            title: entityColumn.labelName,
            record: keyVal
          } as EntityInfoRecordData;
        } else {
          const data = entityColumn.type !== 'fraction'
            ? entityData.attributes[entityColumn.name]
            : [
              entityData.attributes[entityColumn.name],
              entityData.attributes[entityColumn.denominator_name]
            ];
          const convertedText = convertDataToTextForUI(
            data,
            entityColumn,
            dataCache
          );
          return {
            type: 'div',
            key: entityColumn.labelName,
            value: convertedText
          } as EntityInfoRecordData;
        }
    }
  }

  /**
   * Calculate the information on the panel
   * to be displayed from the entity's information.
   *
   * @param entityInfoRecord The information of the form.
   */
  const getPanelData = async (entityInfoRecord: EntityInfoRecord) => {
    // Keep the panel from displaying
    // while the entity's information is being read.
    const panelData: EntityInfoPanelData = {
      title: entityInfoRecord.panelName,
      records: []
    };
    if (!entityData) {
      return panelData;
    }

    const dataCache = await (readDataCache(entityInfoRecord.keyValueRecords, cloudContext));
    for (const entityColumn of entityInfoRecord.keyValueRecords) {
      panelData.records.push(await getEntityInfoRecordData(
        dataCache,
        entityInfoRecord.tableRecordList,
        entityColumn,
      ));
    }

    return panelData;
  }

  /**
   * Calculate the information on the panel
   * to be displayed from the entity's information.
   */
  const getPanelDataList = async () => {
    // Keep the panel from displaying
    // while the entity's information is being read.
    if (!entityData) {
      return [];
    }

    const panelDataList: EntityInfoPanelData[] = [];
    for (const entityInfoRecord of detailTemplate.entityRecords) {
      panelDataList.push(await getPanelData(entityInfoRecord));
    }
    return panelDataList;
  }

  useEffect(() => {
    getPanelDataList().then((data) => {
      setPanelDataList(data);
    });
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [entityData]);

  return {
    panelDataList
  };

}

const InfoTabs = () => {
  const params: {
    cloudContext: string,
    entityId: string,
  } = useParams();
  const { tabTemplate } = useTabTemplate();
  const { t } = useDrupalTranslation();

  if (!tabTemplate) {
    return <></>;
  }

  const url = `/${tabTemplate.cloudServiceProvider}/${params.cloudContext}/${tabTemplate.entityName}/${params.entityId}`;

  return <nav className="tabs">
    <div>
      <ul className="nav nav-tabs" id="tab_wrap">
        <li className="active">
          <Link to={url}
            className="active is-active ripple-effect">
            {Drupal.t('View')}
          </Link>
        </li>
        {
          tabTemplate.tabs.map((tabColumn) => {
            return <li>
              <Link to={`${url}/${tabColumn.name}`}
                className="active is-active ripple-effect">
                {t(tabColumn.labelName)}
              </Link>
            </li>;
          })
        }
      </ul>
    </div>
  </nav>;

}

/**
 * Default page if form data cannot be retrieved.
 */
const DefaultPage = () => {

  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId: string
  } = useParams();

  return <>
    <PageTitle />
    <Breadcrumbs />
    <SideBar />
    <Container fluid className="px-0">
      <Row>
        <Col>
          <span>EntityDetailPage</span><br />
          <span>cloudServiceProvider={params.cloudServiceProvider}</span><br />
          <span>cloudContext={params.cloudContext}</span><br />
          <span>entityName={params.entityName}</span><br />
          <span>entityId={params.entityId}</span><br />
        </Col>
      </Row>
    </Container>
  </>;

}

/**
 * Page of the view of detail info an entity.
 *
 * @param cloudContext The cloud context.
 * @param detailTemplate The detail info template for entity.
 * @param entityId The entity's ID.
 */
const EntityDetailPageImpl = ({ cloudContext, detailTemplate, entityId }: {
  cloudContext: string,
  detailTemplate: EntityInfoTemplate,
  entityId: string,
}) => {

  const { entityData } = useEntityData(
    detailTemplate.cloudServiceProvider,
    detailTemplate.entityName,
    entityId
  );
  const { panelDataList } = usePanelData(cloudContext, detailTemplate, entityId);

  return <>
    <PageTitle />
    <Breadcrumbs />
    <SideBar />
    <InfoTabs />
    <Container fluid className="px-0">
      <Row className="mx-0">
        <Col>
          {
            panelDataList.length > 0
              ? <>
                <DetailActionButtonGroup detailTemplate={detailTemplate} cloudContext={cloudContext}
                  entityData={entityData} entityId={entityId} />
                {panelDataList.map((panelData, index) => {
                  return <EntityInfoPanel key={index} panelData={panelData} />;
                })}
              </>
              : <LoadingSpinner />

          }
        </Col>
      </Row>
    </Container>
  </>;

}

/**
 * Page of the view of detail info an entity.
 */
const EntityDetailPage = () => {

  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityId: string
  } = useParams();
  const { detailTemplate } = useDetailTemplate();

  if (!detailTemplate) {
    return <DefaultPage />;
  }

  return <EntityDetailPageImpl
    cloudContext={params.cloudContext}
    detailTemplate={detailTemplate}
    entityId={params.entityId} />;

}

export default EntityDetailPage;
