import ENTITY_INFO_TAB_LIST from 'constant/templates/entity/entity_info_tab';
import ENTITY_FORM_LIST from 'constant/templates/form/entity_form_template';
import useDrupalJsonApi from 'hooks/drupal_jsonapi';
import useDrupalTranslation from 'hooks/drupal_translation';
import EntityData from 'model/EntityData';
import EntityFormColumn from 'model/EntityFormColumn';
import EntityFormTemplate from 'model/EntityFormTemplate';
import ScriptToken from 'model/ScriptToken';
import EntityFormLabel from 'molecules/EntityFormLabel';
import Breadcrumbs from 'organisms/Breadcrumbs';
import EntityFormPanel from 'organisms/EntityFormPanel';
import PageTitle from 'organisms/PageTitle';
import SideBar from 'organisms/SideBar';
import EntitySubViewPage from 'pages/EntitySubViewPage';
import LoadingDataPage from 'pages/LoadingDataPage';
import OpenStackInstanceConsoleOutputPage from 'pages/openstack/instance/OpenStackInstanceConsoleOutputPage';
import { useEffect, useState } from 'react';
import { Alert, Button, Col, Container, Form, Row } from 'react-bootstrap';
import { Link, useHistory, useParams } from 'react-router-dom';
import { convertFormDataToSendData } from 'service/convert';
import { capitalize, cspToLabel, entityNameToLabel } from 'service/string';

const useFormScript = (cloudContext: string) => {
  const [entity, setEntity] = useState<EntityData | undefined>(undefined);
  const { getJsonData, getEntityListAll } = useDrupalJsonApi();

  /**
 * Split pattern text into separate tokens.
 * ex) 'aaa(bbb, ccc)' =>
 *   [1] type=keyword, value='aaa'
 *   [2] type=bracket, value='('
 *   [3] type=keyword, value='bbb'
 *   [4] type=comma, value=','
 *   [5] type=keyword, value='ccc'
 *   [6] type=bracket, value=')'
 *
 * @param script The pattern text.
 * @returns The separate tokens.
 */
  const convertScriptToToken = (script: string) => {
    const output: ScriptToken[] = [];
    for (let p = 0; p < script.length; p += 1) {
      const character = script[p];
      if (character.match(/[a-zA-Z._]/) !== null) {
        const keyword = script.slice(p).match(/[a-zA-Z._]+/g);
        if (keyword !== null) {
          output.push({ type: 'keyword', value: keyword[0] });
          p += keyword[0].length - 1;
        }
        continue;
      }
      if (character === '"') {
        const keyword = script.slice(p).match(/"[a-zA-Z._]*"/g);
        if (keyword !== null) {
          output.push({ type: 'text', value: keyword[0] });
          p += keyword[0].length - 1;
        }
        continue;
      }
      if (character === '\'') {
        const keyword = script.slice(p).match(/'[a-zA-Z._]*'/g);
        if (keyword !== null) {
          output.push({ type: 'text', value: keyword[0] });
          p += keyword[0].length - 1;
        }
        continue;
      }
      if (character === '(') {
        output.push({ type: 'left-bracket' });
        continue;
      }
      if (character === ')') {
        output.push({ type: 'right-bracket' });
        continue;
      }
      if (character === ',') {
        output.push({ type: 'comma' });
        continue;
      }
    }
    return output;
  }

  /**
   * Join tokens to text.
   *
   * @param tokens The separate tokens.
   */
  const convertTokenToScript = (tokens: ScriptToken[]) => {
    return tokens.map((token) => {
      switch (token.type) {
        case 'text':
          return token.value;
        case 'keyword':
          return token.value;
        case 'left-bracket':
          return '(';
        case 'right-bracket':
          return ')';
        case 'comma':
          return ',';
        case 'null':
          return 'null';
        default:
          return '';
      }
    }).join();
  }

  /**
   * Search for a pattern of tokens that can be processed.
   * @param tokens The separate tokens.
   * @returns If not detected, null is returned.
   */
  const findTokenPattern = (tokens: ScriptToken[]) => {
    // Replace the token type with a symbol for ease of determination.
    const TYPE_THUMBNAIL = {
      'keyword': 'K',
      'text': 'T',
      'right-bracket': 'R',
      'left-bracket': 'L',
      'comma': 'C',
      'null': 'N'
    }
    const thumbnail = tokens.map((t) => {
      return TYPE_THUMBNAIL[t.type];
    }).join('');
    const matches = thumbnail.match(/KL((K|T|N)(C(K|T|N))*)R/g);
    console.group('TokenThumbnail');
    console.log(thumbnail);
    console.log(matches);
    console.groupEnd();
    if (matches === null) {
      return null;
    }
    const match = matches[0];
    const index = thumbnail.indexOf(match);
    return {
      index,
      length: match.length
    };
  }

  /**
   * Script tokens to shrink tokens.
   *
   * @param tokens The separate tokens.
   */
  const convertTokenToToken = async (tokens: ScriptToken[]): Promise<ScriptToken> => {
    const functionName = tokens[0].type === 'keyword' ? tokens[0].value : '';
    if (tokens[0].type !== 'keyword') {
      return { type: 'null' };
    }
    const parameters = tokens.filter(((_, i) => {
      return i >= 2 && i % 2 === 0;
    }));
    console.group('function');
    console.log(functionName);
    console.log(parameters);
    console.groupEnd();
    switch (functionName) {
      case 'relationship': {
        if (entity === undefined) {
          return { type: 'null' };
        }
        const relationshipKey1 = parameters[0].type === 'keyword' || parameters[0].type === 'text'
          ? parameters[0].value : null;
        const relationshipKey2 = parameters[1].type === 'keyword' || parameters[1].type === 'text'
          ? parameters[1].value : null;
        if (relationshipKey1 !== null && relationshipKey2 !== null && typeof entity.relationships !== 'string') {
          const rawValue = entity.relationships[relationshipKey1];
          if (rawValue !== undefined && rawValue.data !== null) {
            const relationshipsData = Array.isArray(rawValue.data) ? rawValue.data[0] : rawValue.data;
            return { type: 'text', value: `${
              ('meta' in relationshipsData)
                ? relationshipsData.meta[relationshipKey2]
                : ''
            }` };
          }
        }
        return { type: 'null' };
      }
      case 'getValue': {
        const entityTypeId = parameters[0].type === 'keyword' || parameters[0].type === 'text'
          ? parameters[0].value : null;
        const entityValue1 = parameters[1].type === 'keyword' || parameters[1].type === 'text'
          ? parameters[1].value : null;
        const entityKey1 = parameters[2].type === 'keyword' || parameters[2].type === 'text'
          ? parameters[2].value : null;
        const entityKey2 = parameters[3].type === 'keyword' || parameters[3].type === 'text'
          ? parameters[3].value : null;
        if (entityTypeId !== null && entityValue1 !== null
          && entityKey1 !== null && entityKey2 !== null) {
          const param: Record<string, any> = {};
          param[`filter[${entityKey1}]`] = entityValue1;
          param['filter[cloud_context]'] = cloudContext;
          const entityList = await getEntityListAll(entityTypeId, param);
          if (entityList.length >= 1) {
            return { type: 'text', value: entityList[0].attributes[entityKey2] };
          } else {
            return { type: 'null' };
          }
        } else {
          return { type: 'null' };
        }
      }
      case 'tryNotNull': {
        const value = parameters[0].type === 'keyword' || parameters[0].type === 'text'
          ? parameters[0].value : null;
        const defaultValue = parameters[1].type === 'keyword' || parameters[1].type === 'text'
          ? parameters[1].value : null;
        const value2 = value !== null ? value : defaultValue;
        return value2 !== null
          ? { type: 'text', value: value2 }
          : { type: 'null' };
      }
    }

    return { type: 'null' };
  }

  /**
   * Convert default value script to string literal by interpreting them.
   * ex.) "foo(bar, baz)" => "xyz"
   *
   * @param script The default value script.
   */
  const calcScript = async (script: string) => {
    // Convert default value script to script tokens.
    const tokens = convertScriptToToken(script);
    console.group('Token');
    console.log(tokens);
    console.groupEnd();

    // Convert script tokens to string;
    let tokensTemp: ScriptToken[] = JSON.parse(JSON.stringify(tokens));
    while (true) {
      // Search for a pattern of tokens that can be processed.
      const result = findTokenPattern(tokensTemp);
      console.group('TokenPattern');
      console.log(result);
      console.groupEnd();
      if (result === null) {
        break;
      }

      // Interpret token, convert to string, and replace tokensTemp.
      const result2 = await convertTokenToToken(tokensTemp.slice(result.index, result.index + result.length));
      tokensTemp = [
        ...tokensTemp.slice(0, result.index),
        result2,
        ...tokensTemp.slice(result.index + result.length)
      ];
      console.group('NewToken');
      console.log(tokensTemp);
      console.groupEnd();
    }
    return convertTokenToScript(tokensTemp);
  }

  /**
   * Convert default value scripts to string literals by interpreting them.
   * ex.) "{foo}{bar}{baz}" => "xxxyyyzzz"
   *
   * @param script The default value script.
   */
  const convertDefaultValueTextFromScript = async (script: string) => {
    // If script_string don't contain '{}', it return script_string.
    const patterns = script.match(/\{[^{}]+?\}/g);
    if (patterns === null || patterns.length === 0) {
      return script;
    }

    // Replace '{keyword}' to 'result_string' from script_string.
    console.group('Script');
    console.log(script);
    console.log(patterns);
    let output = script;
    for (const pattern of patterns) {
      const result = await calcScript(pattern.slice(1, pattern.length - 1));
      output = output.replace(pattern, result);
    }
    console.log(output);
    console.groupEnd();

    return output;
  }

  /**
   * Calculate information to determine default value in form.
   * @param keyValueRecord KeyValueRecord.
   * @param cloudContext The cloud context.
   * @param entityId The entity ID.
   */
  const convertToFormInfoList = async (
    keyValueRecord: EntityFormColumn,
    cloudContext: string,
    entityId: string
  ) => {
    if (keyValueRecord.type === 'time') {
      return [
        { key: keyValueRecord.hourName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[0] },
        { key: keyValueRecord.minuteName, type: keyValueRecord.type, defaultValue: keyValueRecord.defaultValue[1] }
      ];
    }

    if (keyValueRecord.type === 'file') {
      return [{
        key: keyValueRecord.name,
        type: keyValueRecord.type,
        defaultValue: null
      }];
    }

    if (!('defaultValue' in keyValueRecord)) {
      return [{
        key: keyValueRecord.name,
        type: keyValueRecord.type,
        defaultValue: undefined
      }];
    }

    if ('defaultValueUrl' in keyValueRecord && keyValueRecord.defaultValueUrl !== undefined) {
      const replacedUrl = keyValueRecord.defaultValueUrl
        .replace('{cloud_context}', cloudContext)
        .replace('{entity_id}', entityId);
      if (!replacedUrl.endsWith('/')) {
        const jsonData = await getJsonData<string>(replacedUrl);
        if (jsonData.length > 0) {
          return [{
            key: keyValueRecord.name,
            type: keyValueRecord.type,
            defaultValue: jsonData,
          }];
        }
      }
    }

    if (Array.isArray(keyValueRecord.defaultValue)) {
      const defaultValueArray = [];
      for (const record of keyValueRecord.defaultValue) {
        if (typeof record === 'string') {
          defaultValueArray.push(await convertDefaultValueTextFromScript(record));
        } else {
          defaultValueArray.push(record);
        }
      }
      return [{
        key: keyValueRecord.name,
        type: keyValueRecord.type,
        defaultValue: defaultValueArray
      }];
    }

    return [{
      key: keyValueRecord.name,
      type: keyValueRecord.type,
      defaultValue: typeof keyValueRecord.defaultValue === 'string'
        ? await convertDefaultValueTextFromScript(keyValueRecord.defaultValue)
        : keyValueRecord.defaultValue
    }];
  }

  return {
    entity,
    setEntity,
    convertToFormInfoList
  };
}

/**
 * Custom Hooks for detail page tab's template.
 */
const useTabTemplate = () => {

  const params: {
    cloudServiceProvider: string,
    entityName: string,
  } = useParams();

  const tabTemplateList = ENTITY_INFO_TAB_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && template.entityName === params.entityName;
  });

  return {
    tabTemplate: tabTemplateList.length >= 1
      ? tabTemplateList[0] : undefined
  };

}

const InfoTabs = () => {
  const params: {
    cloudContext: string,
    entityId: string,
    action: string,
  } = useParams();
  const { tabTemplate } = useTabTemplate();
  const { t } = useDrupalTranslation();

  if (!tabTemplate) {
    return <></>;
  }

  const url = `/${tabTemplate.cloudServiceProvider}/${params.cloudContext}/${tabTemplate.entityName}/${params.entityId}`;

  return <nav className="tabs">
    <div>
      <ul className="nav nav-tabs" id="tab_wrap">
        <li>
          <Link to={url}
            className="active is-active ripple-effect">
            {Drupal.t('View')}
          </Link>
        </li>
        {
          tabTemplate.tabs.map((tabColumn) => {
            return <li className={tabColumn.name === params.action ? 'active' : ''} key={tabColumn.name}>
              <Link to={`${url}/${tabColumn.name}`}
                className="active is-active ripple-effect">
                {t(tabColumn.labelName)}
              </Link>
            </li>;
          })
        }
      </ul>
    </div>
  </nav>;

}

const EntityXxudPageImpl = ({ cloudContext, entityFormTemplate, entityName, entityId }: {
  cloudContext: string,
  entityFormTemplate: EntityFormTemplate,
  entityName: string,
  entityId: string
}) => {
  const [formData, setFormData] = useState<Record<string, any>>({});
  const [loadingFlg, setLoadingFlg] = useState(true);
  const [entityData, setEntityData] = useState<EntityData | undefined>();
  const [errorMessage, setErrorMessage] = useState('');
  const { entity, setEntity, convertToFormInfoList } = useFormScript(cloudContext);
  const { getJsonData } = useDrupalJsonApi();
  const history = useHistory();
  const { t } = useDrupalTranslation();

  // Initialize form data.
  useEffect(() => {
    setLoadingFlg(true);
    initEntity().then(() => {
      setLoadingFlg(false);
    });
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);

  useEffect(() => {
    if (entityData !== undefined) {
      setEntity(entityData);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [entityData]);

  useEffect(() => {
    if (entity !== undefined) {
      initForm(entity);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [entity]);

  useEffect(() => {
    console.group('FormData');
    console.log(formData);
    console.groupEnd();
  }, [formData]);

  const initEntity = async () => {
    let url = '';
    if (entityName !== "cloud_launch_template") {
      const entityTypeId = `${entityFormTemplate.cloudServiceProvider}_${entityName}`;
      url = `/jsonapi/${entityTypeId}/${entityTypeId}?filter[drupal_internal__id]=${entityId}`;
    } else {
      url = `/jsonapi/cloud_launch_template/${entityFormTemplate.cloudServiceProvider}?filter[drupal_internal__id]=${entityId}`;
    }
    const rawEntityData = await getJsonData<{ data: EntityData[] }>(url);
    if (rawEntityData.data.length >= 1) {
      setEntityData(rawEntityData.data[0]);
    }
  }

  const initForm = async (entityData: EntityData) => {
    // Create a list of key strings and default values.
    const formInfoList: { key: string, type: string, defaultValue: any }[] = [];
    for (const entityRecord of entityFormTemplate.entityRecords) {
      // Process only forms contained in Panel.
      if (entityRecord.type !== 'panel') {
        continue;
      }

      // Load the form inside each Panel.
      for (const keyValueRecord of entityRecord.keyValueRecords) {
        const newFormInfoList = await convertToFormInfoList(
          keyValueRecord,
          cloudContext,
          entityId
        );

        for (const formInfo of newFormInfoList) {
          formInfoList.push(formInfo);
        }
      }
    }

    // Load entity data.
    const attributes = entityData.attributes;
    const newFormData: Record<string, any> = {};
    for (const formInfo of formInfoList) {
      // Handling of types that are not entered in the form, such as 'cost' types.
      if (formInfo.defaultValue === undefined) {
        continue;
      }

      // Processing when the value corresponding to a key cannot be read.
      if (!(formInfo.key in attributes)) {
        newFormData[formInfo.key] = formInfo.defaultValue;
        continue;
      }
      const value = attributes[formInfo.key];
      if (value === null || (Array.isArray(value) && value.length === 0)) {
        newFormData[formInfo.key] = formInfo.defaultValue;
        continue;
      }


      // Processing in 'multi-select' type forms.
      if (formInfo.type === 'multi-select' || formInfo.type === 'array') {
        newFormData[formInfo.key] = typeof value === 'string'
          ? [value] : value;
        continue;
      }

      // Processing in 'radio' type forms.
      if (typeof value === 'boolean' && formInfo.type === 'radio') {
        newFormData[formInfo.key] = value ? '1' : '0';
        continue;
      }

      // Other type.
      newFormData[formInfo.key] = value;
    }

    console.group('NewFormData');
    console.log(newFormData);
    console.groupEnd();
    setFormData(newFormData);
  }

  const save = async () => {
    const url = entityName !== 'cloud_launch_template'
      ? `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.cloudServiceProvider}_${entityName}/${entityId}/${entityFormTemplate.actionType}`
      : `/cloud_dashboard/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}/${entityId}/${entityFormTemplate.actionType}`;

    const response = await fetch(url, {
      method: 'POST',
      body: convertFormDataToSendData(formData, entityFormTemplate),
    });
    const actionTypeC = capitalize(entityFormTemplate.actionType);
    if (!response.ok) {
      console.group(`${actionTypeC} Request`);
      console.log('Result: Failed');
      console.log('Response:');
      console.error(await response.json());
      console.groupEnd();
      setErrorMessage(`${actionTypeC} failed.`);
    } else {
      console.group(`${actionTypeC} Request`);
      console.log('Result: Success');
      console.log('Response:');
      console.log(await response.json());
      console.groupEnd();

      const listPageUrl = entityName === 'cloud_launch_template'
        ? `/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/server_template`
        : `/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityName}`;
      history.push(listPageUrl);
    }
  }

  const saveButtonLabel = entityFormTemplate.submitButtonLabel
    ? entityFormTemplate.submitButtonLabel
    : 'Save';

  return loadingFlg
    ? <LoadingDataPage />
    : <>
      <PageTitle />
      <Breadcrumbs />
      <SideBar />
      <InfoTabs />
      <Container fluid className="px-0">
        <Row className="mx-0">
          <Col>
            <Form>
              {
                errorMessage !== ''
                  ? <Alert variant="danger">{`Error : ${errorMessage}`}</Alert>
                  : <></>
              }
              <Form.Group className="mb-4">
                <Link className="btn btn-primary" to={`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.entityName}`}>
                  <span className="glyphicon glyphicon-th-list"></span> {
                    Drupal.t('List @cloudServiceProvider @entityName', {
                      '@cloudServiceProvider': cspToLabel(entityFormTemplate.cloudServiceProvider),
                      '@entityName': entityNameToLabel(entityFormTemplate.entityName),
                    })
                  }
                </Link>
              </Form.Group>
              {
                entityFormTemplate.entityRecords.map((entityRecord) => {
                  switch (entityRecord.type) {
                    case 'panel':
                      return <EntityFormPanel
                        cloudServiceProvider={entityFormTemplate.cloudServiceProvider}
                        cloudContext={cloudContext}
                        entityName={entityName}
                        actionType={entityFormTemplate.actionType}
                        entityRecord={entityRecord}
                        formData={formData}
                        setFormData={setFormData} />;
                    case 'label':
                      return <EntityFormLabel label={entityRecord.text} entityName={entityName} entityData={entityData} />;
                    default:
                      return <></>;
                  }
                })
              }
              {
                saveButtonLabel !== 'Delete'
                  ? <Button onClick={save}>
                    <span className="glyphicon glyphicon-ok"></span>{' ' + t(saveButtonLabel)}
                  </Button>
                  : <Button onClick={save} variant="danger">
                    <span className="glyphicon glyphicon-remove"></span>{' ' + t(saveButtonLabel)}
                  </Button>
              }
              {
                entityFormTemplate.actionType === 'edit'
                  ? <Link className="btn btn-danger" to={`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.entityName}/${entityId}/delete`}>
                    <span className="glyphicon glyphicon-trash"></span> {Drupal.t('Delete')}
                  </Link>
                  : <Link className="btn btn-danger" to={`/${entityFormTemplate.cloudServiceProvider}/${cloudContext}/${entityFormTemplate.entityName}/${entityId}`}>
                    <span className="glyphicon glyphicon-remove"></span> {Drupal.t('Cancel')}
                  </Link>
              }
            </Form>
          </Col>
        </Row>
      </Container>
    </>;
}

const EntityXxudPage = () => {
  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
    entityId: string,
    action: string
  } = useParams();

  // OpenStack Instance console output page
  if (params.cloudServiceProvider === 'openstack'
    && params.entityName === 'instance'
    && params.action === 'console_output') {
    return <OpenStackInstanceConsoleOutputPage />;
  }

  // Entity Update/Delete page
  const entityFormTemplate = ENTITY_FORM_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && (template.entityName === params.entityName || template.entityName === '')
      && template.actionType === params.action
  });

  if (entityFormTemplate.length >= 1) {
    return <EntityXxudPageImpl
      cloudContext={params.cloudContext}
      entityFormTemplate={entityFormTemplate[0]}
      entityName={params.entityName}
      entityId={params.entityId} />;
  }

  // Entity SubView Page
  const tabTemplateList = ENTITY_INFO_TAB_LIST.filter((template) => {
    return template.cloudServiceProvider === params.cloudServiceProvider
      && template.entityName === params.entityName;
  });

  if (tabTemplateList.length >= 1) {
    const subTemplatelist = tabTemplateList[0].tabs.filter((template) => {
      return template.name === params.action;
    });
    if (subTemplatelist.length >= 1) {
      return <EntitySubViewPage
        cloudContext={params.cloudContext}
        tabTemplate={tabTemplateList[0]}
        entityId={params.entityId}
        subType={params.action} />;
    }
  }

  // Default Page
  return <>
    <PageTitle />
    <Breadcrumbs />
    <SideBar />
    <Container fluid className="px-0">
      <Row>
        <Col>
          <span>EntityCxudPage</span><br />
          <span>cloudServiceProvider={params.cloudServiceProvider}</span><br />
          <span>cloudContext={params.cloudContext}</span><br />
          <span>entityName={params.entityName}</span><br />
          <span>entityId={params.entityId}</span><br />
          <span>action={params.action}</span>
        </Col>
      </Row>
    </Container>
  </>;
}

export default EntityXxudPage;
