import ErrorAlert from 'atoms/ErrorAlert';
import EntityFormRecord from 'model/EntityFormRecord';
import Breadcrumbs from 'organisms/Breadcrumbs';
import EntityFormPanel from 'organisms/EntityFormPanel';
import PageTitle from 'organisms/PageTitle';
import SideBar from 'organisms/SideBar';
import { useEffect, useState } from 'react';
import { Button, Col, Container, Form, Row } from 'react-bootstrap';
import { useHistory, useParams } from 'react-router-dom';

/**
 * Information on the first step form.
 */
const preStackPanel: EntityFormRecord = {
  type: 'panel',
  panelName: 'Stack',
  keyValueRecords: [
    {
      type: 'select-local', labelName: 'Template source', name: 'template_source', defaultValue: '', value: [
        { name: 'file', labelName: 'File' },
        { name: 'data', labelName: 'Direct Input' },
        { name: 'url', labelName: 'URL' },
      ], required: true
    },
    { type: 'file', labelName: 'Template File', name: 'template_file' },
    { type: 'textarea', labelName: 'Template Data', name: 'template_data', defaultValue: '' },
    { type: 'default', labelName: 'Template URL', name: 'template_url', defaultValue: '' },
    {
      type: 'select-local', labelName: 'Environment source', name: 'environment_source', defaultValue: '', value: [
        { name: 'file', labelName: 'File' },
        { name: 'data', labelName: 'Direct Input' },
      ]
    },
    { type: 'file', labelName: 'Environment File', name: 'environment_file' },
    { type: 'textarea', labelName: 'Environment Data', name: 'environment_data', defaultValue: '' },
  ]
}

/**
 * Information on the second step form (Stack panel).
 */
const stackPanel: EntityFormRecord = {
  type: 'panel',
  panelName: 'Stack',
  keyValueRecords: [
    { type: 'default', labelName: 'Name', name: 'name', defaultValue: '' },
    { type: 'number', labelName: 'Creation timeout (minutes)', name: 'timeout_mins', defaultValue: 60 },
    { type: 'boolean', labelName: 'Rollback on failure', name: 'rollback', defaultValue: false },
  ]
}

/**
 * Type to define the template parameter form in the second step of the form.
 */
type FormParameters = {
  parameters: Record<string, {
    default: any,
    description: string,
    type: string,
  }>
}

/**
 * Result of the first step form submission.
 */
type PreFormResult = {
  templateData: FormParameters,
  template: string,
  environment: string,
};

/**
 * Convert the results of the first step form submission
 * into a template parameter form definition for the second step form.
 *
 * @param template The template data.
 */
const convertData = (template: FormParameters) => {
  const output: EntityFormRecord = {
    type: 'panel',
    panelName: 'Template parameters',
    keyValueRecords: []
  };
  const defaultValue: Record<string, any> = {};

  for (const formKey of Object.keys(template.parameters)) {
    const formParameter = template.parameters[formKey];
    switch (formParameter.type) {
      case 'number':
        output.keyValueRecords.push({
          type: 'number',
          labelName: formKey,
          name: 'parameter_' + formKey,
          defaultValue: formParameter.default ?? 0
        });
        defaultValue['parameter_' + formKey] = formParameter.default ?? 0;
        break;

      case 'boolean':
        output.keyValueRecords.push({
          type: 'boolean',
          labelName: formKey,
          name: 'parameter_' + formKey,
          defaultValue: ['t', 'true', 'on', 'y', 'yes', '1'].includes(`${formParameter.default}`.toLowerCase())
        });
        defaultValue['parameter_' + formKey] = ['t', 'true', 'on', 'y', 'yes', '1'].includes(`${formParameter.default}`.toLowerCase());
        break;

      case 'json':
        output.keyValueRecords.push({
          type: 'default',
          labelName: formKey,
          name: 'parameter_' + formKey,
          defaultValue: JSON.stringify(formParameter.default ?? '')
        });
        defaultValue['parameter_' + formKey] = JSON.stringify(formParameter.default ?? '');
        break;

      case 'comma_delimited_list':
        output.keyValueRecords.push({
          type: 'default',
          labelName: formKey,
          name: 'parameter_' + formKey,
          defaultValue: JSON.stringify(formParameter.default ?? '')
        });
        defaultValue['parameter_' + formKey] = JSON.stringify(formParameter.default ?? '');
        break;

      default:
        output.keyValueRecords.push({
          type: 'default',
          labelName: formKey,
          name: 'parameter_' + formKey,
          defaultValue: formParameter.default ?? ''
        });
        defaultValue['parameter_' + formKey] = formParameter.default ?? '';
        break;
    }
  }

  return {
    parameterPanel: output as EntityFormRecord,
    parameterDefaultValue: defaultValue
  };
}

const OpenStackHeatCreatePage = () => {

  const params: {
    cloudServiceProvider: string,
    cloudContext: string,
    entityName: string,
  } = useParams();
  const [formMode, setFormMode] = useState<'pre' | 'main'>('pre');
  const [preFormData, setPreFormData] = useState<Record<string, any>>({
    template_source: '',
    template_file: null,
    template_data: '',
    template_url: '',
    environment_source: '',
    environment_file: null,
    environment_data: '',
  });
  const [parameterPanel, setParameterPanel] = useState<EntityFormRecord>({
    type: 'panel',
    panelName: 'Template parameters',
    keyValueRecords: []
  });
  const [formData, setFormData] = useState<Record<string, any>>({
    name: '',
    timeout_mins: 60,
    rollback: false,
  });
  const [errorMessage, setErrorMessage] = useState('');
  const history = useHistory();

  /**
   * to Next Page.
   */
  const toNext = async () => {
    const url = `/cloud_dashboard/openstack/${params.cloudContext}/openstack_stack/pre_create`;

    const formDataForSubmit = new FormData();
    formDataForSubmit.append('template_source', preFormData.template_source);
    formDataForSubmit.append('template_file', preFormData.template_file);
    formDataForSubmit.append('template_data', preFormData.template_data);
    formDataForSubmit.append('template_url', preFormData.template_url);
    formDataForSubmit.append('environment_source', preFormData.environment_source);
    formDataForSubmit.append('environment_file', preFormData.environment_file);
    formDataForSubmit.append('environment_data', preFormData.environment_data);

    const response = await fetch(url, {
      method: 'POST',
      body: formDataForSubmit,
    });
    if (!response.ok) {
      console.group('Create Request');
      console.log('Result: Failed');
      console.log('Response:');
      const errorResponse: {
        response: string,
        reason: string,
      } = await response.json();
      console.error(errorResponse);
      console.groupEnd();
      setErrorMessage(errorResponse.reason);
    } else {
      console.group('Create Request');
      console.log('Result: Success');
      console.log('Response:');
      const dataResponse: PreFormResult = await response.json();
      console.log(dataResponse);
      console.groupEnd();
      const { parameterPanel, parameterDefaultValue } = convertData(dataResponse.templateData);
      setFormData({
        ...formData,
        ...parameterDefaultValue,
        'template': dataResponse.template,
        'environment': dataResponse.environment,
      });
      setParameterPanel(parameterPanel);
      setErrorMessage('');
      setFormMode('main');
    }
  }

  /**
   * Save data.
   */
  const save = async () => {
    const url = `/cloud_dashboard/openstack/${params.cloudContext}/openstack_stack/create`;

    const formDataForSubmit = new FormData();
    for (const key in formData) {
      formDataForSubmit.append(key, `${formData[key]}`);
    }
    formDataForSubmit.append('template_url', preFormData['template_url']);

    const response = await fetch(url, {
      method: 'POST',
      body: formDataForSubmit,
    });
    if (!response.ok) {
      console.group('Create Request');
      console.log('Result: Failed');
      console.log('Response:');
      console.error(await response.json());
      console.groupEnd();
      setErrorMessage('Create failed.');
    } else {
      console.group('Create Request');
      console.log('Result: Success');
      console.log('Response:');
      console.log(await response.json());
      console.groupEnd();
      history.push(`/openstack/${params.cloudContext}/stack`);
    }
  }

  useEffect(() => {
    console.group('PreFormData');
    console.log(preFormData);
    console.groupEnd();
  }, [preFormData]);

  useEffect(() => {
    console.group('FormData');
    console.log(formData);
    console.groupEnd();
  }, [formData]);

  if (formMode === 'pre') {
    return <>
      <PageTitle action="create" />
      <Breadcrumbs action="create" />
      <SideBar />
      <Container fluid className="px-0">
        <Row className="mx-0">
          <Col>
            <Form>
              <ErrorAlert message={errorMessage} />
              <EntityFormPanel
                cloudServiceProvider="openstack"
                cloudContext={params.cloudContext}
                entityName="stack"
                actionType="create"
                entityRecord={preStackPanel}
                formData={preFormData}
                setFormData={setPreFormData} />
              <Button className="ripple-effect" onClick={toNext}>
                {Drupal.t('Next')}
              </Button>
            </Form>
          </Col>
        </Row>
      </Container>
    </>;
  }
  return <>
    <PageTitle action="create" />
    <Breadcrumbs action="create" />
    <SideBar />
    <Container fluid className="px-0">
      <Row className="mx-0">
        <Col>
          <Form>
            <ErrorAlert message={errorMessage} />
            <EntityFormPanel
              cloudServiceProvider="openstack"
              cloudContext={params.cloudContext}
              entityName="stack"
              actionType="create"
              entityRecord={stackPanel}
              formData={formData}
              setFormData={setFormData} />
            <EntityFormPanel
              cloudServiceProvider="openstack"
              cloudContext={params.cloudContext}
              entityName="stack"
              actionType="create"
              entityRecord={parameterPanel}
              formData={formData}
              setFormData={setFormData} />
            <Button className="ripple-effect" onClick={save}>
              {Drupal.t('Save')}
            </Button>
          </Form>
        </Col>
      </Row>
    </Container>
  </>;

}

export default OpenStackHeatCreatePage;
