import { VMWARE_OS_DICT } from 'constant/others/other';
import CloudContext from 'model/CloudContext';
import DataRecord from 'model/DataRecord';
import EntityColumn from 'model/EntityColumn';
import EntityData from 'model/EntityData';
import EntityFormColumn from 'model/EntityFormColumn';
import EntityFormTemplate from 'model/EntityFormTemplate';
import { convertDateString } from 'service/string';

/**
 * Record<string, any> to FormData.
 *
 * @param formData Record<string, any>.
 * @param entityFormTemplate Template of entity form.
 *
 * @returns FormData type.
 */
export const convertFormDataToSendData = (formData: Record<string, any>, entityFormTemplate: EntityFormTemplate) => {
  const fileTypeRecordList = [];
  for (const record1 of entityFormTemplate.entityRecords) {
    if (record1.type === 'panel') {
      for (const record2 of record1.keyValueRecords) {
        if (record2.type === 'file') {
          fileTypeRecordList.push(record2.name);
        }
      }
    }
  }

  const sendData = new FormData();
  for (const pair of Object.entries(formData)) {
    switch (typeof pair[1]) {
      case 'object':
        sendData.append(
          pair[0],
          fileTypeRecordList.includes(pair[0])
            ? pair[1]
            : JSON.stringify(pair[1])
        );
        break;
      default:
        sendData.append(pair[0], `${pair[1]}`);
        break;
    }
  }
  return sendData;
}

/**
 * Round number method.
 * @param value number
 * @param digit digit
 * @returns fixed value string
 */
const roundNumber = (value: number, digit: number) => {
  // This code means "x = 10 ^ digit".
  let x = 1.0;
  for (let i = 0; i < digit; i += 1) {
    x *= 10;
  }

  // Use guard clause.
  if (Math.floor(value) !== value) {
    return `${Math.round(value * x) / x}`;
  }

  let textBuffer = `${value}`;
  if (digit !== 0) {
    textBuffer += '.';
    for (let i = 0; i < digit; i += 1) {
      textBuffer += '0';
    }
  }
  return textBuffer;
};

const convertKeyValueCrLfData = (data: string) => {
  if (data === '{  }') {
    return '';
  }

  const records = data.split('\n');
  const CONVERT_KEY_VALUE_CRLF_DICT: Record<string, string> = {
    'on_demand_hourly: ': 'On-demand hourly ($): ',
    'on_demand_daily: ': 'On-demand daily ($): ',
    'on_demand_monthly: ': 'On-demand monthly ($): ',
    'on_demand_yearly: ': 'On-demand yearly ($): ',
    'ri_one_year: ': 'RI 1 year ($): ',
    'ri_three_year: ': 'RI 3 year ($): ',
    'cpu: ': 'CPU (Usage): ',
    'memory: !!float ': 'Memory (Usage): ',
    'memory: ': 'Memory (Usage): ',
    'pod_allocation: ': 'Pods (Allocation): ',
    'cpu_capacity: !!float ': 'CPU (Capacity): ',
    'memory_capacity: !!float ': 'Memory (Capacity): ',
    'pod_capacity: ': 'Pods (Capacity): ',
  };
  const fixedRecords: string[] = [];
  for (const record of records) {
    if (record === '') {
      continue;
    }
    let fixedRecord = record;
    for (const key in CONVERT_KEY_VALUE_CRLF_DICT) {
      if (record.includes(key)) {
        const value = record.replace(key, '');
        const valueFloat = parseFloat(value);
        if (key.includes('!!float') || key.includes('memory: ')) {
          if (valueFloat >= 1024 * 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024 * 1024), 2)}Gi`;
          } else if (valueFloat >= 1024 * 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024 * 1024), 2)}Mi`;
          } else if (valueFloat >= 1024) {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat / (1024), 2)}Ki`;
          } else {
            fixedRecord = `${CONVERT_KEY_VALUE_CRLF_DICT[key]}${roundNumber(valueFloat, 2)}`;
          }
        } else {
          fixedRecord = CONVERT_KEY_VALUE_CRLF_DICT[key].includes('Pods ')
            ? CONVERT_KEY_VALUE_CRLF_DICT[key] + value
            : CONVERT_KEY_VALUE_CRLF_DICT[key] + roundNumber(valueFloat, 2);
        }
        break;
      }
    }
    fixedRecords.push(fixedRecord);
  }
  return fixedRecords.sort((a, b) => {
    return a < b ? -1 : a > b ? 1 : 0;
  }).join('\n');
}

/**
 * Converter of entity's data for UI.
 * @param data Entity's data
 * @param ec Information of Entity Column
 * @param dataCache Data cache for 'join' type
 */
export const convertDataToTextForUI = (data: any, ec: EntityColumn, dataCache: { [key: string]: EntityData[] }): string => {
  // Null check.
  if (data === null) {
    return '';
  }

  switch (ec.type) {
    case 'datetime':
      return convertDateString(data);
    case 'cpu':
      return roundNumber(data, 2);
    case 'memory': {
      if (data >= 1024 * 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024 * 1024), 2)}Gi`;
      }
      if (data >= 1024 * 1024) {
        return `${roundNumber(data / (1024 * 1024), 2)}Mi`;
      }
      if (data >= 1024) {
        return `${roundNumber(data / (1024), 2)}Ki`;
      }
      return roundNumber(data, 2);
    }
    case 'key-value': {
      let records: string[] = [];
      for (const record of data) {
        records.push(`${record['item_key']}:${record['item_value']}`);
      }
      return records.join(', ');
    }
    case 'cost':
      return `$${data}`;
    case 'boolean':
      return data ? ec.value[0] : ec.value[1];
    case 'array':
      if (!Array.isArray(data)) {
        return data;
      }
      if (data.length === 0) {
        return '';
      }
      return data.map((r: any) => `${r}`).join(', ');
    case 'join': {
      const arrayData = Array.isArray(data)
        ? data : [data];
      const output: string[] = [];
      for (const datum of arrayData) {
        let flg = false;
        for (const record of dataCache[ec.info.entityTypeId]) {
          const recordData: EntityData = record;
          if (recordData.attributes[ec.info.keyColumn] === datum) {
            output.push(`${recordData.attributes[ec.info.valueColumn]} (${datum})`);
            flg = true;
            break;
          }
        }
        if (!flg) {
          output.push(`${datum}`);
        }
      }
      return output.join(', ');
    }
    case 'key-value-crlf':
      return convertKeyValueCrLfData(data);
    case 'conditions':
      return data !== ec.value[0] ? data : ec.value[1];
    case 'cidr':
      return data.length > 0 ? data[0]['cidr'] : '';
    case 'number':
      return data.toString();
    case 'vmware_os':
      // source: VmwareVmGuestOsDataProvider.php
      return data in VMWARE_OS_DICT ? VMWARE_OS_DICT[data] : data;
    case 'link':
      return JSON.stringify({
        'url': ec.url,
        'label': ec.label,
        'type': ec.type,
      });
    case 'fraction': {
      const numeratorValue = data[0];
      const denominatorValue = data[1];
      if (typeof numeratorValue !== 'number' || typeof denominatorValue !== 'number') {
        return '-1 (N/A)';
      }
      if (denominatorValue === 0) {
        return '-1 (N/A)';
      }
      if (denominatorValue < 0) {
        return `${numeratorValue} (No limit)`;
      }
      const percent = 100.0 * numeratorValue / denominatorValue;
      return `${numeratorValue} of ${denominatorValue} (${Math.round(percent * 10) / 10}%)`;
    }
    case 'select-local':
      const filteredItems = ec.value.filter((record) => {
        return record.name === `${data}`;
      });
      return filteredItems.length >= 1
        ? filteredItems[0].labelName
        : `${data}`;
    default:
      return data;
  }
};

/**
 * Convert EntityData's list to DataRecord's list.
 *
 * @param entityTypeId ID of entity type.
 * @param entityDataList List of EntityData.
 * @param entityColumnList List of EntityColumn.
 * @param cloudContextList List of CloudContext.
 * @param dataCache Data cache of EntityData.
 * @returns List of DataRecord.
 */
export const convertEntityDataToDataRecord = (
  entityTypeId: string,
  entityDataList: EntityData[],
  entityColumnList: EntityColumn[],
  cloudContextList: CloudContext[],
  dataCache: { [key: string]: EntityData[] }) => {
  const newDataRecordList: DataRecord[] = [];

  // Convert EntityData to DataRecord.
  for (const rawRecord of entityDataList) {
    const dataRecord: DataRecord = {
      id: `${rawRecord.attributes['drupal_internal__id']}`,
      entityTypeId,
      value: {},
      cloudContext: rawRecord.attributes['cloud_context'],
    };

    // Read data column by launchTemplateColumn and convert it.
    for (const launchTemplateColumn of entityColumnList) {
      if (!(launchTemplateColumn.name in rawRecord.attributes)) {
        /**
         * Handling of cases where data cannot be retrieved
         * from rawRecorrd.attributes.
         */
        if (launchTemplateColumn.type !== 'relationship') {
          continue;
        }
        if (typeof rawRecord.relationships === 'string') {
          continue;
        }
        if (!(launchTemplateColumn.name in rawRecord.relationships)) {
          continue;
        }

        // Read data column by launchTemplateColumn.
        const rawValue = rawRecord.relationships[launchTemplateColumn.name];
        const relationshipsData = Array.isArray(rawValue.data)
          ? rawValue.data[0] : rawValue.data;
        if (!('meta' in relationshipsData)) {
          dataRecord.value[launchTemplateColumn.name] = '';
          continue;
        }

        const entityId = relationshipsData.meta[launchTemplateColumn.info.keyColumn1];
        const entityList = dataCache[launchTemplateColumn.info.entityTypeId];
        const selectedEntity = entityList.filter((e) => {
          return e.attributes[launchTemplateColumn.info.keyColumn2] === entityId;
        });
        dataRecord.value[launchTemplateColumn.name] = selectedEntity.length >= 1
          ? selectedEntity[0].attributes[launchTemplateColumn.info.valueColumn]
          : '';
        continue;
      }

      // Read data column by launchTemplateColumn.
      const rawValue = launchTemplateColumn.type !== 'fraction'
        ? rawRecord.attributes[launchTemplateColumn.name]
        : [
          rawRecord.attributes[launchTemplateColumn.name],
          rawRecord.attributes[launchTemplateColumn.denominator_name],
        ];

      // Convert as appropriate.
      if (launchTemplateColumn.name !== 'cloud_context') {
        // If it is not 'cloud_context', convert it by convertDataToTextForUI().
        const newValue = convertDataToTextForUI(rawValue, launchTemplateColumn, dataCache);
        dataRecord.value[launchTemplateColumn.name] = newValue;
        continue;
      }

      // Search for possible candidates.
      const candidate = cloudContextList.filter((r) => {
        return r.name !== 'ALL' && r.name === rawValue;
      });

      // Branching depending on the number of candidates.
      dataRecord.value[launchTemplateColumn.name] = candidate.length >= 1
        ? candidate[0].labelName
        : rawValue;
    }
    // Write it out.
    newDataRecordList.push(dataRecord);
  }
  return newDataRecordList;
};

export const getOperationsImpl = (entityTypeId: string, attributes: Record<string, any>) => {
  switch (entityTypeId) {
    case 'aws_cloud_instance': {
      // TODO: support 'Associate Elastic IP' menu.
      const list: string[] = ['Create Image'];
      if (attributes['instance_state'] === 'running') {
        list.push('Stop');
        list.push('Reboot');
      } else if (attributes['instance_state'] === 'stopped') {
        list.push('Start');
      }
      list.push('Associate Elastic IP');
      return list;
    }
    case 'aws_cloud_volume': {
      const list: string[] = ['Create snapshot'];
      if (attributes['state'] === 'available') {
        list.push('Attach');
      } else if (attributes['state'] === 'in-use') {
        list.push('Detach');
      }
      return list;
    }
    case 'aws_cloud_elastic_ip':
      return attributes['association_id'] === ''
        || attributes['association_id'] === null
        || attributes['association_id'] === undefined
        ? ['Associate']
        : ['Disassociate'];
    case 'aws_cloud_snapshot':
      return ['Create Volume'];
    case 'aws_cloud_security_group':
      return ['Copy'];
    case 'aws_cloud_vpc_peering_connection':
      return attributes['status_code'] === 'pending-acceptance'
        ? ['Accept'] : [];
    case 'aws_cloud_internet_gateway':
      return attributes['vpc_id'] === ''
        || attributes['vpc_id'] === null
        ? ['Attach']
        : ['Detach'];
    case 'k8s_deployment':
      return ['Scale'];
    case 'cloud_launch_template': {
      const list: string[] = [];
      if (attributes['field_workflow_status'] === 'Approved') {
        list.push('Launch');
      }
      list.push('Copy');
      if (attributes['field_workflow_status'] === 'Review') {
        list.push('Approve');
      }
      if (attributes['field_workflow_status'] === 'Draft') {
        list.push('Review');
      }
      return list;
    }
    case 'cloud_project':
      return ['Launch', 'Copy'];
    case 'k8s_cost_store':
      return ['Copy'];
    case 'k8s_namespace_resource_store':
      return ['Copy'];
    case 'k8s_node_resource_store':
      return ['Copy'];
    case 'k8s_pod_resource_store':
      return ['Copy'];
    case 'openstack_instance': {
      const list: string[] = [];
      if (attributes['instance_state'] === 'running') {
        list.push('Stop');
        list.push('Reboot');
        list.push('Console output');
      } else if (attributes['instance_state'] === 'stopped') {
        list.push('Start');
      }
      list.push('Attach interface')
      list.push('Detach interface')
      return list;
    }
    case 'openstack_floating_ip':
      return attributes['association_id'] === ''
        || attributes['association_id'] === null
        || attributes['association_id'] === undefined
        ? ['Associate']
        : ['Disassociate'];
    case 'openstack_volume':
      return attributes['attachment_information'] === ''
        || attributes['attachment_information'] === null
        || attributes['attachment_information'] === undefined
        ? ['Attach']
        : ['Detach'];
    case 'openstack_stack':
      return ['Check', 'Suspend', 'Resume'];
    case 'openstack_user':
      return ['Change password'];
    case 'vmware_vm':
      const list: string[] = [];
      if (attributes['power_state'] === 'POWERED_OFF') {
        list.push('Start');
      }
      if (attributes['power_state'] === 'POWERED_ON') {
        list.push('Stop');
        list.push('Suspend');
        list.push('Reboot');
      }
      if (attributes['power_state'] === 'SUSPENDED') {
        list.push('Start');
        list.push('Stop');
      }
      return list;
  }
  return [];
}

/**
 * Find and return the list of items to be added to Operation links.
 *
 * @param dataRecord Record of data.
 * @returns The list of items.
 */
export const getOperations = (dataRecord: DataRecord): string[] => {
  return getOperationsImpl(dataRecord.entityTypeId, dataRecord.value);
}

/**
 * Determines whether to display the form or not.
 *
 * Some pages require fine control over the display of forms. Therefore, this
 * function determines whether it should display the form or not.
 * @param cloudServiceProvider The cloud service provider.
 * @param entityName The entity's type name.
 * @param actionType The CxUD action Type.
 * @param entityFormColumn The entity form column.
 * @param formData The data of form values.
 *
 * @returns If TRUE, the block is displayed.
 */
export const showEntityFormBlockFlg = (
  cloudServiceProvider: string,
  entityName: string,
  actionType: string,
  entityFormColumn: EntityFormColumn,
  formData: Record<string, any>
) => {
  // Associate AWS Cloud Elastic IP
  if (actionType === 'associate'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'elastic_ip') {
    if (entityFormColumn.type === 'select') {
      switch (entityFormColumn.name) {
        case 'instance_id':
          return formData['resource_type'] === 'instance';
        case 'instance_private_ip':
          return formData['resource_type'] === 'instance';
        case 'network_interface_id':
          return formData['resource_type'] === 'network_interface';
        case 'network_interface_private_ip':
          return formData['resource_type'] === 'network_interface';
      }
    }
  }

  // Create K8s deployment
  if (actionType === 'create'
    && cloudServiceProvider === 'k8s'
    && entityName === 'deployment') {
    switch (entityFormColumn.labelName) {
      case 'Detail':
        return formData['time_scheduler_option'] === 'cloud_orchestrator_scheduler';
      case 'YAML URL':
        return formData['time_scheduler_option'] === 'cronjob_scheduler';
      case 'Scheduler Use Type':
        return formData['enableTimeScheduler'];
      case 'Start-up Time':
        return formData['enableTimeScheduler'];
      case 'Stop Time':
        return formData['enableTimeScheduler'];
    }
  }

  // Create K8s pod
  if (actionType === 'create'
    && cloudServiceProvider === 'k8s'
    && entityName === 'pod') {
    switch (entityFormColumn.labelName) {
      case 'Detail':
        return formData['time_scheduler_option'] === 'cloud_orchestrator_scheduler';
      case 'YAML URL':
        return formData['time_scheduler_option'] === 'cronjob_scheduler';
      case 'Scheduler Use Type':
        return formData['enableTimeScheduler'];
      case 'Start-up Time':
        return formData['enableTimeScheduler'];
      case 'Stop Time':
        return formData['enableTimeScheduler'];
    }
  }

  // Associate OpenStack Floating IP
  if (actionType === 'associate'
    && cloudServiceProvider === 'openstack'
    && entityName === 'floating_ip') {
    if (entityFormColumn.type === 'select') {
      switch (entityFormColumn.name) {
        case 'instance_id':
          return formData['resource_type'] === 'instance';
        case 'instance_private_ip':
          return formData['resource_type'] === 'instance';
        case 'network_id':
          return formData['resource_type'] === 'network';
      }
    }
  }

  // Create AWS Cloud VPC Peering Connection
  if (actionType === 'create'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'vpc_peering_connection') {

    if (entityFormColumn.type === 'default'
      || entityFormColumn.type === 'select'
      || entityFormColumn.type === 'hidden') {

      // If the regions and account_ids are different, display the textfield
      // `accepter_vpc_id_text` so users can enter a freeform VPC ID belonging
      // to a different region.
      if (entityFormColumn.name === 'accepter_vpc_id') {
        if (formData['accepter_region'] !== formData['default_region'] ||
          formData['accepter_account_id'] !== formData['default_account_id']
        ) {
          formData['accepter_vpc_id'] = '';
          return false;
        }
      }

      if (entityFormColumn.name === 'accepter_vpc_id_text') {
        if (formData['accepter_region'] === formData['default_region'] &&
          formData['accepter_account_id'] === formData['default_account_id']
        ) {
          formData['accepter_vpc_id_text'] = '';
          return false;
        }
      }
    }
  }

  // Edit AWS Cloud Transit Gateway
  if (actionType === 'edit'
    && cloudServiceProvider === 'aws_cloud'
    && entityName === 'transit_gateway'
  ) {
    if (entityFormColumn.type === 'boolean' || entityFormColumn.type === 'select') {
      switch (entityFormColumn.name) {
        case 'association_default_route_table_id':
          return formData['default_route_table_association'];
        case 'propagation_default_route_table_id':
          return formData['default_route_table_propagation'];
      }
    }
  }

  // Create/Edit Openstack Subnet
  if ((actionType === 'create' || actionType === 'edit')
    && cloudServiceProvider === 'openstack'
    && entityName === 'subnet') {
    if (entityFormColumn.type === 'default') {
      if (entityFormColumn.name === 'gateway_ip') {
        return formData['disable_gateway'] === false;
      }
    }
  }

  // Create/Edit Openstack Router
  if ((actionType === 'create' || actionType === 'edit')
    && cloudServiceProvider === 'openstack'
    && entityName === 'router') {
    if (entityFormColumn.type === 'boolean') {
      if (entityFormColumn.name === 'external_gateway_enable_snat') {
        return formData['external_gateway_network_id'] !== '';
      }
    }
  }

  // Create/Edit Openstack Port
  if ((actionType === 'create' || actionType === 'edit')
    && cloudServiceProvider === 'openstack'
    && entityName === 'port') {
    if (entityFormColumn.type === 'default'
      || entityFormColumn.type === 'select'
      || entityFormColumn.type === 'multi-select'
      || entityFormColumn.type === 'item-array') {
      switch (entityFormColumn.name) {
        case 'subnet':
          return formData['ip_address_or_subnet'] === 'subnet';
        case 'fixed_ips':
          return formData['ip_address_or_subnet'] === 'fixed_ip';
        case 'security_groups':
          return formData['port_security_enabled'] === true;
        case 'allowed_address_pairs':
          return formData['port_security_enabled'] === true;
      }
    }
  }

  // Create Openstack Heat
  if (actionType === 'create'
    && cloudServiceProvider === 'openstack'
    && entityName === 'stack') {
    if (entityFormColumn.type === 'file'
      || entityFormColumn.type === 'default'
      || entityFormColumn.type === 'textarea') {
      switch (entityFormColumn.name) {
        case 'template_file':
          return formData['template_source'] === 'file';
        case 'template_data':
          return formData['template_source'] === 'data';
        case 'template_url':
          return formData['template_source'] === 'url';
        case 'environment_file':
          return formData['environment_source'] === 'file';
        case 'environment_data':
          return formData['environment_source'] === 'data';
      }
    }
  }

  // Preview Openstack Heat
  if (actionType === 'preview'
    && cloudServiceProvider === 'openstack'
    && entityName === 'stack') {
    if (entityFormColumn.type === 'file'
      || entityFormColumn.type === 'default'
      || entityFormColumn.type === 'textarea') {
      switch (entityFormColumn.name) {
        case 'template_file':
          return formData['template_source'] === 'file';
        case 'template_data':
          return formData['template_source'] === 'data';
        case 'template_url':
          return formData['template_source'] === 'url';
        case 'environment_file':
          return formData['environment_source'] === 'file';
        case 'environment_data':
          return formData['environment_source'] === 'data';
      }
    }
  }

  return true;
}
