<?php

namespace Drupal\cloud\Controller;

use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;

/**
 * Controller responsible for "update" URLs.
 *
 * This class is mainly responsible for updating the entities from URLs.
 */
class ApiController extends ControllerBase implements ApiControllerInterface {

  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  /**
   * The entity used by this form.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface|\Drupal\Core\Entity\RevisionLogInterface
   */
  private $entity;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * ApiController constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager
  ) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Dependency Injection.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Instance of ContainerInterface.
   *
   * @return ApiController
   *   return created object.
   */
  public static function create(ContainerInterface $container): ApiController {
    return new static(
      $container->get('entity_type.manager'),
    );
  }

  /**
   * Gets the entity of this form.
   *
   * Provided by \Drupal\Core\Entity\EntityForm.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   The entity.
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function operateEntity(Request $request, string $cloud_service_provider, string $cloud_context, string $entity_type_id, string $entity_id, string $command): JsonResponse {
    // Create an instance of the entity.
    // The process splits between adding a new entity and when reading an
    // existing entity.
    /** @var \Drupal\cloud\Entity\CloudLaunchTemplateInterface $entity */
    $entity = $command !== 'create'
      ? $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->load($entity_id)
      : $this->entityTypeManager
        ->getStorage($entity_type_id)
        ->create([
          'cloud_context' => $cloud_context,
        ]);
    if (empty($entity)) {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'The entity has already been deleted.',
      ], 404);
    }

    if ($command !== 'delete') {
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'Not Found',
      ], 404);
    }

    try {
      $this->entity = $entity;
      $entity->delete();

      $this->messenger->addStatus($this->getDeletionMessage());
      $this->logDeletionMessage();

      // Clear block and menu cache.
      $this->clearCacheValues($this->entity->getCacheTags());
      return new JsonResponse([
        'result' => 'OK',
      ], 200);
    }
    catch (\Exception $e) {
      $this->handleException($e);
      return new JsonResponse([
        'result' => 'NG',
        'reason' => 'Internal Server Error',
      ], 500);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getImageOptions($cloud_service_provider, $cloud_context): JsonResponse {
    /** @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface[] $images */
    $images = $this->entityTypeManager
      ->getStorage("{$cloud_service_provider}_image")
      ->loadByProperties([
        'cloud_context' => $cloud_context,
      ]);

    $options = [];
    foreach ($images as $image) {
      $options[$image->getImageId()] = sprintf('%s (%s)',
        $image->getName(),
        $image->getImageId()
      );
    }

    natcasesort($options);

    $output = [];
    foreach ($options ?: [] as $value => $label) {
      $output[] = [
        'value' => $value,
        'label' => $label,
      ];
    }
    return new JsonResponse($output, 200);
  }

}
