<?php

namespace Drupal\Tests\cloud\Traits;

use Drupal\cloud\Entity\CloudLaunchTemplateInterface;

/**
 * The trait creating form data for cloud server template testing.
 */
trait CloudLaunchTemplateTrait {

  /**
   * Check the status for List and canonical view.
   *
   * @param string $list_url
   *   The URL of the cloud launch template list.
   * @param string $view_url
   *   The URL of the cloud launch template canonical view.
   * @param string $workflow_status
   *   The Value of workflow status to be checked.
   */
  private function checkPageTextContains($list_url, $view_url, $workflow_status): void {
    // Validate the display of the list page.
    $this->drupalGet($list_url);
    $this->assertSession()->pageTextContains($workflow_status);
    switch ($workflow_status) {
      case CloudLaunchTemplateInterface::DRAFT:
        $this->assertSession()->linkExistsExact($this->t('Review'));
        $this->assertSession()->linkNotExistsExact($this->t('Approve'));
        $this->assertSession()->linkNotExistsExact($this->t('Launch'));
        break;

      case CloudLaunchTemplateInterface::REVIEW:
        $this->assertSession()->linkNotExistsExact($this->t('Review'));
        $this->assertSession()->linkNotExistsExact($this->t('Launch'));
        break;

      case CloudLaunchTemplateInterface::APPROVED:
        $this->assertSession()->linkNotExistsExact($this->t('Review'));
        $this->assertSession()->linkNotExistsExact($this->t('Approve'));
        $this->assertSession()->linkExistsExact($this->t('Launch'));
        break;
    }
    // Validate the display of the canonical page.
    $this->drupalGet($view_url);
    $this->assertSession()->pageTextContains($workflow_status);
    switch ($workflow_status) {
      case CloudLaunchTemplateInterface::DRAFT:
        // Launch tab is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Launch'));
        // Approve button is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Approve'));
        // Test case for URL access control.
        $this->assertSession()->linkByHrefExists("$view_url/review");
        $this->assertSession()->linkByHrefNotExists("$view_url/approve");
        $this->assertSession()->linkByHrefNotExists("$view_url/launch");
        // Review button is displayed and can be clicked.
        $this->clickLink('Review', 0);
        // Launch as is not displayed.
        $this->assertSession()->pageTextNotContains($this->t('Launch as'));
        break;

      case CloudLaunchTemplateInterface::REVIEW:
        // Launch tab is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Launch'));
        // Review button is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Review'));
        // Test case for URL access control.
        $this->assertSession()->linkByHrefNotExists("$view_url/review");
        $this->assertSession()->linkByHrefNotExists("$view_url/launch");
        break;

      case CloudLaunchTemplateInterface::APPROVED:
        // Launch tab is not displayed.
        $this->assertSession()->linkExistsExact($this->t('Launch'));
        // Review button is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Review'));
        // Test case for URL access control.
        $this->assertSession()->linkByHrefNotExists("$view_url/review");
        $this->assertSession()->linkByHrefNotExists("$view_url/approve");
        $this->assertSession()->linkByHrefExists("$view_url/launch");
        // Approve button is not displayed.
        $this->assertSession()->linkNotExistsExact($this->t('Approve'));
        break;
    }
  }

  /**
   * Update the cloud launch template to make sure there are no errors.
   *
   * @param string $cloud_context
   *   The cloud context.
   * @param string $test_data
   *   The cloud server template.
   * @param string $type
   *   The type is 'add' or 'edit'.
   * @param string $module_name
   *   The module name.
   * @param int $num
   *   The num is sequential numbering of cloud server template.
   */
  private function updateLaunchTemplate($cloud_context, $test_data, $type, $module_name, $num = 1): void {
    if (empty($cloud_context) || empty($test_data)) {
      $this->assertSame('Either of the two arguments is empty.');
      return;
    }

    $url = '';
    $text_contains = '';
    $t_args = [
      '@type' => 'launch template',
      '%label' => $test_data['name[0][value]'],
    ];

    switch ($type) {
      case CloudLaunchTemplateInterface::OPERATION_ADD:
        $url = "/clouds/design/server_template/$cloud_context/$module_name/add";
        $text_contains = strip_tags($this->t('The @type %label has been created.', $t_args));
        break;

      case CloudLaunchTemplateInterface::OPERATION_EDIT:
        $url = "/clouds/design/server_template/$cloud_context/$num/edit";
        $text_contains = strip_tags($this->t('The @type %label has been updated.', $t_args));
        break;

      default:
        $this->assertSame('There is no corresponding type.');
        return;
    }

    if ($module_name === 'k8s') {
      unset($test_data['field_object']);
    }
    $this->drupalGet($url);
    $this->submitForm(
      $test_data,
      $this->t('Save'));
    $this->assertNoErrorMessage();
    $this->assertSession()->pageTextContains($text_contains);

  }

}
